#include <iostream>
#include <sys/time.h>
#ifdef WITH_SSE
#include <xmmintrin.h>
#endif

using namespace std;

static struct timeval start;
static float buffer[256];

/* Make a 256-sample buffer of 1s and then multiply the
   whole buffer by some value x.
*/
static void
test (float x)
{
	for (int i = 0; i < 256; ++i) {
		buffer[i] = 1;
	}

	for (int i = 0; i < 256; ++i) {
		buffer[i] *= x;
	}
}

/* Get the time in fractional seconds since gettimeofday (&start, 0)
   was last called.
*/
static double
get_dt ()
{
	struct timeval now;
	gettimeofday (&now, 0);
	
	time_t s = now.tv_sec - start.tv_sec;
	suseconds_t u = now.tv_usec - start.tv_usec;
	if (u < 0) {
		s--;
		u += 1e6;
	}

	return s + (double (u) / 1e6);
}

/* Pass ftz as first arg to enable FTZ, daz to enable DAZ
   (only when WITH_SSE is defined)
*/
int main (int argc, char* argv[])
{
	bool ftz = false;
	bool daz = false;
	
	if (argc > 1) {
		if (string (argv[1]) == "ftz") {
			ftz = true;
		} else if (string (argv[1]) == "daz") {
			daz = true;
		}
	}

	/* The number of times that the test will be run */
	int const N = 1e7;

#ifdef WITH_SSE	
	int mxcsr = _mm_getcsr ();

	if (ftz) {
		mxcsr |= (1<<15) | (1<<11);
	}

	if (daz) {
		mxcsr |= (1<<6);
	}

	_mm_setcsr (mxcsr);
#endif

	/* Non-denormal test */

	gettimeofday (&start, 0);

	for (int i = 0; i < N; ++i) {
		test (1);
	}

	double const n = get_dt ();

	/* Denormal test */

	gettimeofday (&start, 0);
	
	for (int i = 0; i < N; ++i) {
		test (1e-39);
	}

	double const s = get_dt ();

#ifdef WITH_SSE	
	cout << "SSE enabled FTZ=" << ftz << " DAZ=" << daz << " ";
#endif

#ifdef WITH_FAST_MATH
	cout << "-ffast-math enabled ";
#endif		

	cout << "subnormal " << (s / n) << " times slower.\n";

	return 0;
}
